package defaultops

import (
	"errors"
	"flow"
	"flow/registry"
	"fmt"
	"math"
	"math/rand"
	"strings"
	"time"
)

// New create a registry
func New() *registry.R {
	r := registry.New()
	// String functions
	registry.Describer(
		r.Add(strings.Split).Inputs("string", "separator"),
		r.Add(strings.Join).Inputs("", "sep"),
		r.Add(strings.Compare, strings.Contains),
		r.Register("Cat", func(a, b string) string { return a + " " + b }),
		r.Register("ToString", func(a interface{}) string { return fmt.Sprint(a) }),
	).Tags("string").Extra("style", registry.M{"color": "#839"})

	// Math functions
	r.Add(
		math.Abs, math.Cos, math.Sin, math.Exp, math.Exp2, math.Tanh, math.Max, math.Min,
	).Tags("math").Extra("style", registry.M{"color": "#386"})

	registry.Describer(
		r.Add(rand.Int, rand.Intn, rand.Float64),
		r.Register("Perm", func(n int) []int {
			if n > 10 { // Limiter for safety
				n = 10
			}
			return rand.Perm(n)
		}),
	).Tags("rand").Extra("style", registry.M{"color": "#486"})

	// Test functions
	r.Add(testErrorPanic, testErrorDelayed, testRandomError).
		Tags("testing-errors")

	registry.Describer(
		r.Register("wait", wait),
		r.Register("waitRandom", waitRandom),
	).Tags("testing-time").Extra("style", map[string]string{"color": "#8a5"})
	return r
}

func wait(data flow.Data, n int) flow.Data {
	time.Sleep(time.Duration(n) * time.Second) // Simulate
	return data
}
func waitRandom(data flow.Data) flow.Data {
	time.Sleep(time.Duration(rand.Intn(10)) * time.Second)
	return data
}

////////////////////
// Testing custom struct operations
/////

////////////////////////
// testOps
////////////////
func testErrorPanic(n int) flow.Data {
	dur := time.Duration(n)
	time.Sleep(dur * time.Second)
	panic("I panicked")
}

func testErrorDelayed(n int) (flow.Data, error) {
	dur := time.Duration(n)
	time.Sleep(dur * time.Second)
	return nil, fmt.Errorf("I got an error %v", dur)
}
func testRandomError(d flow.Data) (flow.Data, error) {
	r := rand.Intn(10)
	if r > 5 {
		return nil, errors.New("I failed on purpose")
	}
	return d, nil
}
