package main

import (
	"flowserver"
	"flowserver/cmd/demo1/assets"
	"flowserver/cmd/demo1/defaultops"
	"flowserver/cmd/demo1/devops"
	"flowserver/cmd/demo1/testops"
	"log"
	"net/http"

	"github.com/gohxs/prettylog"
	"github.com/gohxs/webu"
	"github.com/gohxs/webu/chain"
)

//go:generate go get github.com/gohxs/folder2go
//go:generate folder2go -handler -nobackup static assets assets

func main() {
	prettylog.Global()
	log.Println("Running version:", flowserver.Version)

	addr := ":2015"
	log.Println("Starting server  at:", addr)

	c := chain.New(webu.ChainLogger(prettylog.New("req")))

	mux := http.NewServeMux()
	mux.HandleFunc("/", assets.AssetHandleFunc)

	mux.Handle("/default/", c.Build(flowserver.New(defaultops.New(), "default").ServeHTTP))
	mux.Handle("/devops/", c.Build(flowserver.New(devops.New(), "devops").ServeHTTP))
	mux.Handle("/testops/", c.Build(flowserver.New(testops.New(), "tests").ServeHTTP))

	// Context registry
	http.ListenAndServe(addr, mux)
}
